#include <SPI.h>
#include <SdFat.h>
#include <TinyGPS++.h>
#include <SoftwareSerial.h>
#include "IRLremote.h"
#include "LCD_Functions.h"

//sd card
SdFat sd;
SdFile myFile;

//gps
static const uint8_t RXPin = 4, TXPin = 3;
TinyGPSPlus gps;
SoftwareSerial ss(RXPin, TXPin);
uint8_t gps_fix = 0;

//ir remote
static const uint8_t pinIR = 2;
CNec IRLremote;
char ir_button;

char str_buffer[12];

const char title_1_str[4] PROGMEM = "GPS";
const char title_2_str[7] PROGMEM = "Logger";
const char title_3_str[5] PROGMEM = "v1.0";
const char no_gps_str[6] PROGMEM = "?GPS?";
const char no_sd_str[5] PROGMEM = "?SD?";
const char blank_str[6] PROGMEM = "     ";

const char one_str[3] PROGMEM = "1.";
const char two_str[3] PROGMEM = "2.";
const char three_str[3] PROGMEM = "3.";
const char four_str[3] PROGMEM = "4.";

const char select_str[7] PROGMEM = "Select";
const char mode_str[6] PROGMEM = "Mode:";
const char point_str[6] PROGMEM = "Point";
const char line_str[5] PROGMEM = "Line";
const char area_str[6] PROGMEM = "Area";
const char track_str[6] PROGMEM = "Track";

const char code_str[6] PROGMEM = "Code:";
const char seq_str[6] PROGMEM = "Seq#:";
const char min_s_str[7] PROGMEM = "min/S:";
const char log_str[6] PROGMEM = " Log ";

const char lat_str[5] PROGMEM = "Lat:";
const char long_str[5] PROGMEM = "Lng:";

const char file_name_str[8] PROGMEM = "log.txt";

uint8_t log_mode = 0;
uint8_t display_state = 1;
uint8_t display_state_prev = 0;
uint8_t event_state = 1;
char name_code[6] = "_____";
uint8_t sequence_num = 0;

void setup() {
  // Open serial communications and wait for port to open:
  Serial.begin(9600);
  while (!Serial) {
    ; // wait for serial port to connect. Needed for native USB port only
  }
    
  //lcd
  lcdBegin(); // This will setup our pins, and initialize the LCD
  fill_str_buffer_array(title_1_str);
  setStr(str_buffer, 0, 2, BLACK);
  fill_str_buffer_array(title_2_str);
  setStr(str_buffer, 21, 2, BLACK);
  fill_str_buffer_array(title_3_str);
  setStr(str_buffer, 60, 2, BLACK);
  updateDisplay(); 
  setContrast(60); // Good values range from 40-60
  delay(2000);
  clearDisplay(WHITE);

  //sd card
  if (!sd.begin(10, SPI_HALF_SPEED)){
    // error accessing SD
    fill_str_buffer_array(no_sd_str);
    setStr(str_buffer, 0, 0, BLACK);
    updateDisplay();
  }
  
  //ir remote
  IRLremote.begin(pinIR);

  //gps software serial
  ss.begin(9600);
}

void loop() {

  //update the gps object with data from the GPS module
  while (ss.available() > 0)
    if (gps.encode(ss.read()));

  //gps fix indicator
  if (gps.location.age()<2000){ 
    if (gps_fix == 0 || gps_fix == 1){ //gps fix
      fill_str_buffer_array(blank_str);
      setStr(str_buffer, 55, 0, BLACK);
      updateDisplay();
      gps_fix = 2;
    }
  }
  else{
    if (gps_fix == 0 || gps_fix == 2){ //no gps fix
      fill_str_buffer_array(no_gps_str);
      setStr(str_buffer, 55, 0, BLACK);
      updateDisplay();
      gps_fix = 1;
    }
  }

  if (display_state != display_state_prev){
    display_state_prev = display_state;
    
    if (display_state == 1){ //select mode
      clearDisplayAfter(84*8, WHITE);
      fill_str_buffer_array(select_str);
      setStr(str_buffer, 0, 8, BLACK);
      fill_str_buffer_array(mode_str);
      setStr(str_buffer, 40, 8, BLACK);
    
      // 1. Point
      fill_str_buffer_array(one_str);
      setStr(str_buffer, 5, 16, BLACK);
      fill_str_buffer_array(point_str);
      setStr(str_buffer, 17, 16, BLACK);
    
      // 2. Line
      fill_str_buffer_array(two_str);
      setStr(str_buffer, 5, 24, BLACK);
      fill_str_buffer_array(line_str);
      setStr(str_buffer, 17, 24, BLACK);
    
      // 3. Area
      fill_str_buffer_array(three_str);
      setStr(str_buffer, 5, 32, BLACK);
      fill_str_buffer_array(area_str);
      setStr(str_buffer, 17, 32, BLACK);
    
      // 4. Track
      fill_str_buffer_array(four_str);
      setStr(str_buffer, 5, 40, BLACK);
      fill_str_buffer_array(track_str);
      setStr(str_buffer, 17, 40, BLACK);
      updateDisplay();
    }
    else if (display_state == 2 || display_state == 3){ //mode
      if (log_mode > 0){
        clearDisplayAfter(84*8, WHITE);
        fill_str_buffer_array(mode_str);
        setStr(str_buffer, 0, 8, BLACK);
      
        if (log_mode == 1){
          fill_str_buffer_array(point_str);
          setStr(str_buffer, 35, 8, BLACK);
        }
        else if (log_mode == 2){
          fill_str_buffer_array(line_str);
          setStr(str_buffer, 35, 8, BLACK);
        }
        else if (log_mode == 3){
          fill_str_buffer_array(area_str);
          setStr(str_buffer, 35, 8, BLACK);
        }
        else if (log_mode == 4){
          fill_str_buffer_array(track_str);
          setStr(str_buffer, 35, 8, BLACK);
        }

        if (display_state == 2) { //move cursor to "Code:"
          fill_str_buffer_array(code_str);
          setStr(str_buffer, 0, 16, WHITE);
          fill_str_buffer_array(log_str);
          setStr(str_buffer, 0, 40, BLACK);
        }
        else if (display_state == 3){ //move cursor to "Log"
          fill_str_buffer_array(code_str);
          setStr(str_buffer, 0, 16, BLACK);
          fill_str_buffer_array(log_str);
          setStr(str_buffer, 0, 40, WHITE);
          updateDisplay();
        }

        setStr(name_code, 35, 16, BLACK); //update name code
        
        if (log_mode > 1){
          fill_str_buffer_array(seq_str);
          setStr(str_buffer, 0, 24, BLACK);
          itoa(sequence_num, str_buffer, 10);
          setStr(str_buffer, 35, 24, BLACK);
      
          if (log_mode == 4){
            fill_str_buffer_array(min_s_str);
            setStr(str_buffer, 0, 32, BLACK);
          }
        }
        updateDisplay();
      }
    }
  }

  //if receiving IR signal
  if (IRLremote.available()){
    auto data = IRLremote.read();
    ir_button = ir_remote_hex_to_char(data.command);
    //Serial.println(data.command);

    if (event_state == 1){ // log mode selection   
      if (ir_button == '1' || ir_button == '2' || ir_button == '3' || ir_button == '4'){
        if (ir_button == '1'){
          log_mode = 1;
        }
        else if (ir_button == '2'){
          log_mode = 2;
        }
        else if (ir_button == '3'){
          log_mode = 3;
        }
        else if (ir_button == '4'){
          log_mode = 4;
        }
        sequence_num = 0;
        display_state = 2;
        event_state = 2;
      }   
    }
    
    else if (event_state == 2){ //cursor on "code"
      if (ir_button == '<'){
        display_state = 1;
        event_state = 1;
        log_mode = 0;
      }
      else if (ir_button == '-'){ //move cursor to "log"
        display_state = 3;
        event_state = 3;
      }
      //backspace
      else if (ir_button == '*'){
        for (int i=4; i>=0; i--){
          if (name_code[i] != '_'){
            name_code[i] = '_';
            setStr(name_code, 35, 16, BLACK);
            updateDisplay();
            break;
          }
        }
      }
      //number entry
      else if (ir_button != 'k' && ir_button != 'e' && ir_button != '+' && ir_button != '>'){     
        for (int i=0; i<=4; i++){
          if (name_code[i] == '_'){
            name_code[i] = ir_button;
            setStr(name_code, 35, 16, BLACK);
            updateDisplay();
            break;
          }
        }
      }
    }
    
    else if (event_state == 3){ //cursor on "log"
      if (ir_button == '<'){
        display_state = 1;
        event_state = 1;
        log_mode = 0;
      }
      else if (ir_button == '+'){ //move cursor to "code"
        display_state = 2;
        event_state = 2;
      }
      else if (ir_button == 'k'){ //log gps data to sd card, display gps data on lcd
        if (gps_fix == 2){
          fill_str_buffer_array(file_name_str);
          // write to sd file
          if (!myFile.open(str_buffer, O_RDWR | O_CREAT | O_AT_END)) {
            // error accessing SD
            fill_str_buffer_array(no_sd_str);
            setStr(str_buffer, 0, 0, BLACK);
            updateDisplay();
          } 
          myFile.print(name_code);
          myFile.print(",");
          if (log_mode == 1){ // point
            fill_str_buffer_array(point_str);
            myFile.print(str_buffer);
          }
          else if (log_mode == 2){ //line
            fill_str_buffer_array(line_str);
            myFile.print(str_buffer);
          }
          else if (log_mode == 3){ //line
            fill_str_buffer_array(area_str);
            myFile.print(str_buffer);
          }
          else if (log_mode == 4){ //track
            fill_str_buffer_array(track_str);
            myFile.print(str_buffer);
          }
          myFile.print(",");
          myFile.print(sequence_num);
          myFile.print(",");
          myFile.print(gps.location.lat(), 6);
          myFile.print(",");
          myFile.print(gps.location.lng(), 6);
          myFile.print(",");
          myFile.print(gps.date.month());
          myFile.print("/");
          myFile.print(gps.date.day());
          myFile.print("/");
          myFile.print(gps.date.year());
          myFile.print(",");
          if (gps.time.hour() < 10) myFile.print("0");
          myFile.print(gps.time.hour());
          myFile.print(":");
          if (gps.time.minute() < 10) myFile.print("0");
          myFile.print(gps.time.minute());
          myFile.print(":");
          if (gps.time.second() < 10) myFile.print("0");
          myFile.println(gps.time.second());
          myFile.close();

          //update lcd screen    
          clearDisplayAfter(84*8, WHITE);
          fill_str_buffer_array(seq_str);
          setStr(str_buffer, 0, 8, BLACK);
          itoa(sequence_num, str_buffer, 10);
          setStr(str_buffer, 35, 8, BLACK);
      
          // latitude and longitude
          fill_str_buffer_array(lat_str);
          setStr(str_buffer, 0, 16, BLACK);
          dtostrf(gps.location.lat(), 10, 6, str_buffer);
          setStr(str_buffer, 25, 16, BLACK);          
          fill_str_buffer_array(long_str);
          setStr(str_buffer, 0, 24, BLACK);
          dtostrf(gps.location.lng(), 10, 6, str_buffer);
          setStr(str_buffer, 25, 24, BLACK);
      
          //date mm:dd:yyyy
          if (gps.date.month() < 10){
            setStr("0", 0, 32, BLACK);
            itoa(gps.date.month(), str_buffer, 10);
            setStr(str_buffer, 6, 32, BLACK);
          }
          else {
            itoa(gps.date.month(), str_buffer, 10);
            setStr(str_buffer, 0, 32, BLACK);
          }
          setStr("/", 12, 32, BLACK);
          if (gps.date.day() < 10){
            setStr("0", 18, 32, BLACK);
            itoa(gps.date.day(), str_buffer, 10);
            setStr(str_buffer, 24, 32, BLACK);
          }
          else {
            itoa(gps.date.day(), str_buffer, 10);
            setStr(str_buffer, 18, 32, BLACK);
          }
          setStr("/", 30, 32, BLACK);
          itoa(gps.date.year(), str_buffer, 10);
          setStr(str_buffer, 36, 32, BLACK);
      
          //time hh:mm:ss
          if (gps.time.hour() < 10){
            setStr("0", 0, 40, BLACK);
            itoa(gps.time.hour(), str_buffer, 10);
            setStr(str_buffer, 6, 40, BLACK);
          }
          else {
            itoa(gps.time.hour(), str_buffer, 10);
            setStr(str_buffer, 0, 40, BLACK);
          }
          setStr(":", 12, 40, BLACK);
          if (gps.time.minute() < 10){
            setStr("0", 18, 40, BLACK);
            itoa(gps.time.minute(), str_buffer, 10);
            setStr(str_buffer, 24, 40, BLACK);
          }
          else {
            itoa(gps.time.minute(), str_buffer, 10);
            setStr(str_buffer, 18, 40, BLACK);
          }
          setStr(":", 30, 40, BLACK);
          if (gps.time.second() < 10){
            setStr("0", 36, 40, BLACK);
            itoa(gps.time.second(), str_buffer, 10);
            setStr(str_buffer, 42, 40, BLACK);
          }
          else {
            itoa(gps.time.second(), str_buffer, 10);
            setStr(str_buffer, 36, 40, BLACK);
          }
          updateDisplay();
          
          if (log_mode > 1){
            sequence_num++;
          }

          display_state = 0;
          event_state = 4;
        }
      }
    }
    else if (event_state == 4){ //on gps data screen
      if(ir_button == 'k' || ir_button == '<'){
        display_state = 3;
        event_state = 3;
      }
    }
  }
}

void fill_str_buffer_array(const char* msg){
  int len = strlen_P(msg);
  for (int i=0; i<len; i++){
    str_buffer[i] = pgm_read_byte_near(msg + i);
  }
  str_buffer[len] = '\0'; //clip array to size of msg with null character
}

char ir_remote_hex_to_char(int ir_hex){
  char ir_char = 'e'; //ir_hex = 0x00
  if (ir_hex == 28){
    ir_char = 'k'; //ok
  }
  else if (ir_hex == 24){
    ir_char = '+'; //up
  }
  else if (ir_hex == 82){
    ir_char = '-'; //down
  }
  else if (ir_hex == 8){
    ir_char = '<'; //left
  }
  else if (ir_hex == 90){
    ir_char = '>'; //right
  }         
  else if (ir_hex == 22){
    ir_char = '*';
  }
  else if (ir_hex == 13){
    ir_char = '#';
  }
  else if (ir_hex == 25){
    ir_char = '0';
  }
  else if (ir_hex == 69){
    ir_char = '1';
  }
  else if (ir_hex == 70){
    ir_char = '2';
  }
  else if (ir_hex == 71){
    ir_char = '3';
  }
  else if (ir_hex == 68){
    ir_char = '4';
  }
  else if (ir_hex == 64){
    ir_char = '5';
  }
  else if (ir_hex == 67){
    ir_char = '6';
  }
  else if (ir_hex == 7){
    ir_char = '7';
  }
  else if (ir_hex == 21){
    ir_char = '8';
  }
  else if (ir_hex == 9){
    ir_char = '9';    
  }
  return ir_char;
}
